from typing import Literal, Any, get_args
from jsons import JsonSerializable
from jsons._dump_impl import dump  # I am a bad man
from jsons.exceptions import DeserializationError


def serialize_literal(obj: Any, cls: Literal, **kwargs):
    return dump(obj, type(obj), **kwargs)


def deserialize_literal(
    obj: Any, cls: Literal, *, strictly_equal_literal: bool = False, **kwargs
):
    for literal_value in get_args(cls):
        value_equal = obj == literal_value
        type_equal = type(obj) != type(literal_value)
        if value_equal and (not strictly_equal_literal or type_equal):
            break
    else:
        err_msg = 'Could not match the object "{}" to the Literal: {}'.format(obj, cls)
        raise DeserializationError(err_msg, obj, None)
    return literal_value


class BaseModel(
    JsonSerializable.set_serializer(
        serialize_literal, Literal
    ).set_deserializer(  # type: ignore
        deserialize_literal, Literal
    )
):  # type: ignore
    pass



from typing import Union, Callable, Optional

from discord_interactions_flask import discord_types as types
from discord_interactions_flask import interactions

ButtonFunction = Callable[[interactions.ButtonInteraction], types.InteractionResponse]
TextInputFunction = Callable[
    [interactions.TextInputInteraction], types.InteractionResponse
]
SelectMenuFunction = Callable[
    [interactions.TextInputInteraction], types.InteractionResponse
]


class Button(types.Button):
    def __init__(
        self,
        *args,
        interaction_handler: Optional[ButtonFunction] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._func = interaction_handler

    def handler(self, f: ButtonFunction):
        self._func = f
        return f

    @property
    def interaction_handler(self) -> Optional[ButtonFunction]:
        return self._func

    @interaction_handler.setter
    def interaction_handler(
        self, f: Callable[[interactions.ButtonInteraction], types.InteractionResponse]
    ):
        self.handler(f)

    def __call__(
        self, interaction: interactions.ButtonInteraction
    ) -> types.InteractionResponse:
        if self._func is None:
            raise ValueError("This Button was not assigned a handler")
        return self._func(interaction)


class TextInput(types.TextInput):
    def __init__(
        self,
        *args,
        interaction_handler: Optional[TextInputFunction] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._func = interaction_handler

    def handler(self, f: TextInputFunction):
        self._func = f
        return f

    @property
    def interaction_handler(self) -> Optional[TextInputFunction]:
        return self._func

    @interaction_handler.setter
    def interaction_handler(
        self,
        f: Callable[[interactions.TextInputInteraction], types.InteractionResponse],
    ):
        self.handler(f)

    def __call__(
        self, interaction: interactions.TextInputInteraction
    ) -> types.InteractionResponse:
        if self._func is None:
            raise ValueError("This TextInput was not assigned a handler")
        return self._func(interaction)


class SelectMenu(types.SelectMenu):
    def __init__(
        self,
        *args,
        interaction_handler: Optional[SelectMenuFunction] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._func = interaction_handler

    def handler(self, f: SelectMenuFunction):
        self._func = f
        return f

    @property
    def interaction_handler(self) -> Optional[SelectMenuFunction]:
        return self._func

    @interaction_handler.setter
    def interaction_handler(
        self,
        f: Callable[[interactions.SelectMenuInteraction], types.InteractionResponse],
    ):
        self.handler(f)

    def __call__(
        self, interaction: interactions.SelectMenuInteraction
    ) -> types.InteractionResponse:
        if self._func is None:
            raise ValueError("This SelectMenu was not assigned a handler")
        return self._func(interaction)


Component = Union[Button, TextInput, SelectMenu]



"""Builds :class:`Command` instances from functions."""
from collections.abc import Generator
from contextlib import contextmanager
import inspect
import logging
from typing import Callable, Optional, Union, cast, TYPE_CHECKING


from discord_interactions_flask import discord_types as types
from discord_interactions_flask.command import (
    ChatCommand,
    ChatMetaCommand,
    CommandGroup,
    MessageCommand,
    SubCommand,
    UserCommand,
)
from discord_interactions_flask import interactions

if TYPE_CHECKING:
    from discord_interactions_flask.discord import Discord


logger = logging.getLogger(__name__)

COMMANDS = {
    interactions.ChatInteraction: ChatCommand,
    interactions.UserInteraction: UserCommand,
    interactions.MessageInteraction: MessageCommand,
}

ChatFunction = Callable[[interactions.ChatInteraction], types.InteractionResponse]
UserFunction = Callable[[interactions.UserInteraction], types.InteractionResponse]
MessageFunction = Callable[[interactions.MessageInteraction], types.InteractionResponse]
# There is a term for why I need to specify the type like this instead of `Callable[[types.Interaction], types.InteractionResponse]` being able to apply to all three, I just forget what it is
# The rationale is that `Callable[[types.Interaction], types.InteractionResponse]` means "Function that can take any Interaction and returns an InteractionResponse"
# A function that can only take UserInteractions does not satisfy that
CommandFunction = Union[ChatFunction, UserFunction, MessageFunction]


# TODO: Add description param
class CommandBuilder:
    """Builds :class:`Command` instances from functions. Typically constructed with :meth:`~discord_interactions_flask.discord.Discord.command`."""

    def __init__(
        self,
        discord: 'Discord',
        name: Optional[str],
        description: Optional[str],
        guild_id: Optional[str],
    ):
        """Initialize :class:`CommandBuilder`.

        Args
            discord: A reference to a :class:`discord_interactions_flask.discord.Discord`.
            name: An optional name to use for the command. If not present the command functions name will be used.
        """
        self.discord = discord
        self.name = name
        self.guild_id = guild_id
        self._description = description

        self.context = {}

    def __call__(self, f: CommandFunction):
        """Decorate a command function to create a :class:`Command`.

        .. code-block:: python

            @discord.command()
            def command(interaction: interactions.ChatInteraction) -> types.InteractionResponse:
                return ...

        Args
            f: A callable that takes a :class:`interactions.ChatInteraction`, :class:`interactions.UserInteraction`, or :class:`interactions.MessageInteraction` instance and returns a :class:`types.InteractionResponse`.
        """
        signature = inspect.signature(f)
        param = list(signature.parameters.values())[0]
        if not (command_class := COMMANDS.get(param.annotation)):
            raise ValueError("Command parameter must be a valid Interaction type")

        if self.name:
            name = self.name
        else:
            name = f.__name__
        command = command_class(name, f)
        command.description = self._description
        self.discord.add_command(name, command, self.guild_id)
        return command

    def subcommand(
        self, name: Optional[str] = None
    ) -> Callable[[ChatFunction], SubCommand]:
        """Create a decorator for a command function to create a :class:`Command`. Should only be used within a :func:`group`.

        .. code-block:: python

            @group.subcommand("sub-command")
            def sub_command(interaction: interactions.ChatInteraction) -> types.InteractionResponse:
                return ...

        Args
            name: An optional name to give the subcommand. If not given the functions name is used.
        """

        def outer(f: ChatFunction) -> SubCommand:
            if not name:
                command_name = f.__name__
            else:
                command_name = name

            subcommand = SubCommand(command_name, f)
            self.context[name] = subcommand
            return subcommand

        return outer

    @contextmanager
    def group(self, name: str) -> Generator["CommandBuilder", None, None]:
        """Create a subcommand group.

        .. code-block:: python

            with command.group("group_name") as group:
                group.description = ...
                @group.subcommand()
                ...

        Args
            name: An name for the group.
        """
        group_context = {}
        old_context = self.context
        old_context[name] = group_context
        self.context = group_context
        yield self
        self.context = old_context

    # :grimacing:
    # I don't like this hacky solution around command/group description
    @property
    def description(self):
        """Set the description used for the current command/group."""
        return self.context["__description"]

    @description.setter
    def description(self, value: str):
        self.context["__description"] = value

    @description.deleter
    def description(self):
        del self.context["__description"]

    def __enter__(self) -> "CommandBuilder":
        """Create subcommands and commad grounp.

        .. code-block:: python

            with discord.command("command-name") as command:
                command.description = ...
                with command.group("group") as group:
                    group.description = ...
                    @group.subcommand()
                    ...
                @command.subcommand()
                def subcommand(...):
                    ...
        """
        if not self.name:
            raise ValueError("Commands created with the context manager require a name")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Build the Command containing the groups/subcommands defined with this Builder."""
        if exc_type is None and exc_val is None and exc_tb is None:
            children = {}
            for (name, child) in self.context.items():
                if name.startswith("__"):
                    continue

                if isinstance(child, dict):
                    group = CommandGroup(
                        name=name, description=child.get("__description", "")
                    )
                    for (subname, subcommand) in child.items():
                        if subname.startswith("__"):
                            continue
                        if not isinstance(subcommand, SubCommand):
                            raise ValueError(
                                "Groups may not contain anything other than subcommands"
                            )
                        group.subcommands[subname] = subcommand
                    children[name] = group
                else:
                    children[name] = child

            assert self.name is not None
            meta_command = ChatMetaCommand(
                name=self.name,
                children=children,
                description=self.context.get("__description", ""),
            )
            self.discord.add_command(self.name, meta_command, self.guild_id)



from dataclasses import dataclass, field
from functools import wraps
from typing import Union, Dict, cast
import jsons

from discord_interactions_flask.discord_types import CommandType
from discord_interactions_flask import discord_types as types
from discord_interactions_flask import interactions


class ChatCommand:
    def __init__(self, name: str, func):
        self.name = name
        self.func = func
        self.description = ""
        self.options = []
        wraps(func)(self)

    def add_option(self, option: types.ApplicationCommandOption):
        self.options.append(option)

    def spec(self) -> dict:
        return {
            "name": self.name,
            "type": CommandType.CHAT,
            "description": self.description,
            "options": [
                jsons.dump(o, use_enum_name=False, strip_privates=True)
                for o in self.options
            ],
        }

    def __call__(self, interaction: interactions.ChatInteraction):
        return self.func(interaction)


class SubCommand(ChatCommand):
    def spec(self) -> dict:
        spec = super().spec()
        spec["type"] = types.ApplicationCommandOptionType.SUB_COMMAND
        return spec


@dataclass
class CommandGroup:
    name: str
    description: str = ""
    subcommands: Dict[str, SubCommand] = field(default_factory=dict)

    def spec(self):
        return {
            "name": self.name,
            "description": self.description,
            "type": types.ApplicationCommandOptionType.SUB_COMMAND_GROUP,
            "options": [sc.spec() for (_, sc) in self.subcommands.items()],
        }

    def __call__(
        self, interaction: interactions.ChatInteraction
    ) -> types.InteractionResponse:
        if not interaction.data.options:
            raise ValueError("This should be impossible")

        group_options = interaction.data.options[0]
        if not group_options.options:
            raise ValueError(
                "Expected group to have options to define subcommand options"
            )

        subcommand_data = group_options.options[0]
        subcommand = self.subcommands[subcommand_data.name]
        return subcommand(interaction)


@dataclass
class ChatMetaCommand:
    name: str
    children: Dict[str, Union[CommandGroup, SubCommand]]
    description: str = ""

    def spec(self):
        return {
            "name": self.name,
            "type": CommandType.CHAT,
            "description": self.description,
            "options": [child.spec() for (_, child) in self.children.items()],
        }

    def __call__(
        self, interaction: interactions.ChatInteraction
    ) -> types.InteractionResponse:
        if not interaction.data.options:
            raise ValueError("Expected meta command to have a group or subcommand")

        data = interaction.data.options[0]
        command = self.children[data.name]
        if not command:
            raise ValueError("Subcomand is not part of this ChatMetaCommand")

        return command(interaction)


class UserCommand:
    def __init__(self, name: str, func):
        self.name = name
        self.func = func
        wraps(func)(self)

    def spec(self) -> dict:
        return {
            "name": self.name,
            "type": CommandType.USER,
        }

    def __call__(self, interaction: interactions.UserInteraction):
        return self.func(interaction)


class MessageCommand:
    def __init__(self, name: str, func):
        self.name = name
        self.func = func
        wraps(func)(self)

    def spec(self) -> dict:
        return {
            "name": self.name,
            "type": CommandType.MESSAGE,
        }

    def __call__(self, interaction: interactions.MessageInteraction):
        return self.func(interaction)


Command = Union[ChatCommand, ChatMetaCommand, UserCommand, MessageCommand]



__version__ = "0.2.0"

from discord_interactions_flask.discord import Discord
from discord_interactions_flask.discord_types import (
    InteractionResponse,
)
from discord_interactions_flask.interactions import (
    ChatInteraction,
    UserInteraction,
    MessageInteraction,
)

# TODO: I want to make discord_types an implementation detail that users aren't expected to interact with,
# TODO: I think meta commands might require descriptions
# TODO: Better logging and/or error handling
#       Parse data out of different return signatures? str = regular message with no frills?
# TODO: built-in support for other InteractionCallbackType types
# TODO: Autocomplete support
# TODO: Document discord_types module
# TODO: Do I need a delete method on Discord, just for convenience. It's a PITA to remove a command as-is
#       There's a weird thought in here, keeping a db of commands and having like migrations and shit
# TODO: Your app can have a global CHAT_INPUT and USER command with the same name
#       Do I just decide to not support that?
#       I would have to break out each command type into separate dicts
# TODO: There should be a mechanism to opt-out of automatically init_commands-ing
# TODO: How do modals work?
# TODO: Should we attempt to persist component handlers across invocations? It would likely require pickeling the functions and storing them somewhere
# TODO: Should we also try to offer a more streamlined decorator that automatically parses and creates slash command arguments from type signatures of function params?

# TODO: Better ergonomics on response
# 1. A command should be able to just return a str to make the simple case easy
# 2. Decide how to handle deferred-type responses
# 3. More helpers?



from typing import Optional
from discord_interactions_flask.discord_types import (
    InteractionResponse,
    InteractionCallbackType,
    InteractionCallbackDataMessage,
)


def content_response(content: str, flags: Optional[int] = None) -> InteractionResponse:
    return InteractionResponse(
        type=InteractionCallbackType.CHANNEL_MESSAGE_WITH_SOURCE,
        data=InteractionCallbackDataMessage(content=content, flags=flags),
    )



from dataclasses import dataclass
from typing import Union, List

from discord_interactions_flask.discord_types import (
    Interaction,
    InteractionData,
    MessageComponent,
)

# TODO: These can have their own definition
ChatData = InteractionData
UserData = InteractionData
MessageData = InteractionData
# If we do give them their own definition
# InteractionData = Union[ChatData, UserData, MessageData, ...]


@dataclass
class ChatInteraction(Interaction):
    data: ChatData


@dataclass
class UserInteraction(Interaction):
    data: UserData


@dataclass
class MessageInteraction(Interaction):
    data: MessageData


# TODO: These can have their own definition
ButtonData = MessageComponent
TextInputData = MessageComponent


@dataclass
class SelectMenuData(MessageComponent):
    values: List[str]


@dataclass
class ButtonInteraction(Interaction):
    data: ButtonData


@dataclass
class SelectMenuInteraction(Interaction):
    data: SelectMenuData


@dataclass
class TextInputInteraction(Interaction):
    data: TextInputData


CommandInteraction = Union[ChatInteraction, UserInteraction, MessageInteraction]
ComponentInteraction = Union[
    ButtonInteraction, SelectMenuInteraction, TextInputInteraction
]



"""
Main extension module.

This is the primary interaction point, a user will define their commands using an instance of the :class:`Discord` class.
"""


from collections import defaultdict
import http
import json
import logging
from typing import Optional, Union, Iterable

from flask import Blueprint, Flask, jsonify, request

import urllib3

from discord_interactions_flask import discord_types as types
from discord_interactions_flask.verify import verify_key_decorator
from discord_interactions_flask.command_builder import CommandBuilder
from discord_interactions_flask.command import Command
from discord_interactions_flask import errors
from discord_interactions_flask.interactions import (
    ChatInteraction,
    UserInteraction,
    MessageInteraction,
    ButtonInteraction,
    SelectMenuInteraction,
    TextInputInteraction,
    ComponentInteraction,
    CommandInteraction,
)
from discord_interactions_flask import helpers
from discord_interactions_flask import components

logger = logging.getLogger(__name__)

GLOBAL_URL_TEMPLATE = "https://discord.com/api/v10/applications/%s/commands"
GUILD_URL_TEMPLATE = "https://discord.com/api/v10/applications/%s/guilds/%s/commands"
OAUTH_ENDPOINT = "https://discord.com/api/v10/oauth2/token"


def _missing_component_handler(
    _: ComponentInteraction,
) -> types.InteractionResponse:
    resp = helpers.content_response(
        ":warning: Sorry, that component has expired! :warning:",
        flags=types.MessageFlags.EPHEMERAL,
    )
    return resp


def _missing_command_handler(
    _: CommandInteraction,
) -> types.InteractionResponse:
    resp = helpers.content_response(
        ":warning: Sorry, I don't understand that command! :warning:",
        flags=types.MessageFlags.EPHEMERAL,
    )
    return resp


class Discord:
    """Creates :class:`Command` s, submits them to the configured Discord application, and recieves the HTTP requests."""

    def __init__(
        self,
        app: Optional[Flask] = None,
        missing_command_handler=_missing_command_handler,
        missing_component_handler=_missing_component_handler,
    ):
        """Initialzation."""

        # TODO: Would be nice if I didn't have to maintain two separate dicts of commands
        self.commands: defaultdict[Optional[str], dict[str, Command]] = defaultdict(dict)

        self.runtime_commands: dict[str, Command] = {}

        # TODO: evict old handlers
        self.component_handlers: defaultdict[
            str, dict[str, components.Component]
        ] = defaultdict(dict)

        self.missing_command_handler = missing_command_handler
        self.missing_component_handler = missing_component_handler
        self.http = urllib3.PoolManager(headers={"Content-Type": "application/json"})
        self.public_key = None
        if app:
            self.init_app(app)

    def command(
            self, name: Optional[str] = None, description: Optional[str] = None, guild_id: Optional[str] = None
    ) -> CommandBuilder:
        """Define a new command.

        Type hints on the argument types are used to build the correct Command.
        Check the examples to see the options.

        .. code-block:: python

            @discord.command()
            def command(interaction: interactions.ChatInteraction) -> types.InteractionResponse:
                return ...

        Args
            name: An optional name to give the command. If not given the function name will be used.
            description: An optional description to give the command. If not given the name will be used.
            guild_id: An optional guild_id. If given the command will be created in just that guild.

        Returns
            A :class:`CommandBuilder` instance that can be used as a decorator or context manager.
        """
        return CommandBuilder(self, name, description, guild_id)

    def _handle_response(
        self, interaction: types.Interaction, response: types.InteractionResponse
    ):
        if response.data and response.data.components:
            for row in response.data.components:
                for component in row.components:
                    self.component_handlers[interaction.id][
                        component.custom_id
                    ] = component

        return jsonify(
            response.dump(
                use_enum_name=False, strip_privates=True, strip_properties=True
            )
        )

    def init_app(self, app: Flask) -> None:
        """Initialize the :class:`Flask` instance with all the commands defined on this :class:`Discord` instance.

        Args
            app: A :class:`Flask` instance. Must have the `DISCORD_PUBLIC_KEY`, `DISCORD_CLIENT_ID`, and `DISCORD_CLIENT_SECRET` configuration keys defined.
        """
        self.public_key = app.config.setdefault("DISCORD_PUBLIC_KEY", "")
        if not self.public_key:
            raise ValueError("You must define a DISCORD_PUBLIC_KEY configuration value")

        # TODO: I'd like to move the bp out of init_app, the verify_key_decorator makes it awkward since with it I need the public key at bp creation time
        interactions_bp = Blueprint("interactions", __name__, url_prefix="/discord")

        @interactions_bp.post("/interactions")
        @verify_key_decorator(self.public_key)
        def interactions():
            payload = request.json
            assert payload is not None

            match payload["type"]:
                case types.InteractionType.PING:
                    return jsonify(type=types.InteractionCallbackType.PONG)
                case types.InteractionType.APPLICATION_COMMAND:
                    command_interaction: Union[
                        ChatInteraction, UserInteraction, MessageInteraction
                    ]
                    match payload["data"]["type"]:
                        case types.CommandType.CHAT:
                            command_interaction = ChatInteraction.load(payload)
                        case types.CommandType.USER:
                            command_interaction = UserInteraction.load(payload)
                        case types.CommandType.MESSAGE:
                            command_interaction = MessageInteraction.load(payload)
                        case _:
                            raise errors.DiscordInteractionsFlaskError(
                                "Discord sent an invalid command type - %s"
                                % payload["data"]["type"]
                            )

                    handler = self.runtime_commands.get(command_interaction.data.id)
                    if not handler:
                        result = self.missing_command_handler(command_interaction)
                    else:
                        result = handler(command_interaction)

                    return self._handle_response(command_interaction, result)
                case types.InteractionType.MESSAGE_COMPONENT:
                    component_interaction: Union[
                        ButtonInteraction, SelectMenuInteraction, TextInputInteraction
                    ]
                    match payload["data"]["component_type"]:
                        case types.ComponentType.BUTTON:
                            component_interaction = ButtonInteraction.load(payload)
                        case types.ComponentType.SELECT_MENU:
                            component_interaction = SelectMenuInteraction.load(payload)
                        case types.ComponentType.TEXT_INPUT:
                            component_interaction = TextInputInteraction.load(payload)
                        case _:
                            raise errors.DiscordInteractionsFlaskError(
                                "Discord sent an invalid component type - %s"
                                % payload["data"]["type"]
                            )

                    # It's interesting that typing errors have pushed me to explicitly define my invariants
                    assert component_interaction.message is not None
                    assert component_interaction.message.interaction is not None

                    handlers = self.component_handlers.get(
                        component_interaction.message.interaction.id
                    )
                    if not handlers or component_interaction.data.custom_id not in handlers:
                        result = self.missing_component_handler(component_interaction)
                    else:
                        result = handlers[component_interaction.data.custom_id](component_interaction)

                    return self._handle_response(component_interaction, result)
                case _:
                    print("OTHER")
                    print(payload)
                    return ("", http.HTTPStatus.NO_CONTENT)

        app.register_blueprint(interactions_bp)
        self.init_commands(app)

    def _refresh_token(self):
        """Refresh the OAuth2 client credentials used to interact with the Discord API. Typically not something a user is expected to need."""
        r = self.http.request(
            "POST",
            OAUTH_ENDPOINT,
            fields={
                "grant_type": "client_credentials",
                "scope": "applications.commands.update",
            },
            headers=urllib3.util.make_headers(
                basic_auth=f"{self.client_id}:{self.client_secret}"
            ),
            encode_multipart=False,
        )
        token = json.loads(r.data.decode("utf-8"))
        self.http.headers[
            "Authorization"
        ] = f"{token['token_type']} {token['access_token']}"

    def _create_commands(
        self, commands: Iterable[Command], guild_id: Optional[str] = None
    ) -> list[types.ApplicationCommand]:
        """Push the list of :class:`Command` to the Discord application indicated by the DISCORD_CLIENT_ID key in bulk.

        Args
            command: The command the create.
            guild_id: if not present, the command will be created as a global one. Otherwise it will be created for the specifiied guild.
        """
        if guild_id:
            url = GUILD_URL_TEMPLATE % (self.client_id, guild_id)
        else:
            url = GLOBAL_URL_TEMPLATE % self.client_id

        resp = self.http.request(
            "PUT", url, body=json.dumps([command.spec() for command in commands]).encode("utf-8")
        )
        if resp.status == http.HTTPStatus.OK:
            return [types.ApplicationCommand.load(payload) for payload in json.loads(resp.data.decode("utf-8"))]
        else:
            raise errors.DiscordApiError(resp.data.decode("utf-8"))

    def _create_command(
        self, command: Command, guild_id: Optional[str] = None
    ) -> types.ApplicationCommand:
        """Push the :class:`Command` to the Discord application indicated by the DISCORD_CLIENT_ID key.

        Args
            command: The command the create.
            guild_id: if not present, the command will be created as a global one. Otherwise it will be created for the specifiied guild.
        """
        if guild_id:
            url = GUILD_URL_TEMPLATE % (self.client_id, guild_id)
        else:
            url = GLOBAL_URL_TEMPLATE % self.client_id

        resp = self.http.request(
            "POST", url, body=json.dumps(command.spec()).encode("utf-8")
        )
        if resp.status == http.HTTPStatus.OK or resp.status == http.HTTPStatus.CREATED:
            interaction_payload = json.loads(resp.data.decode("utf-8"))
            return types.ApplicationCommand.load(interaction_payload)
        else:
            raise errors.DiscordApiError(resp.data.decode("utf-8"))

    def add_command(
        self, name: str, command: Command, guild_id: Optional[str] = None
    ) -> None:
        self.commands[guild_id][name] = command

        # If we're already initialized, we send the command to discord right away
        if self.public_key is not None:
            interaction = self._create_command(command, guild_id)
            self.runtime_commands[interaction.id] = command

    def remove_command(
        self, interaction_id: str, guild_id: Optional[str] = None
    ) -> None:
        if guild_id:
            url = GUILD_URL_TEMPLATE % (self.client_id, guild_id)
        else:
            url = GLOBAL_URL_TEMPLATE % self.client_id

        url += f"/{interaction_id}"
        del self.runtime_commands[interaction_id]

        resp = self.http.request("DELETE", url)
        if resp.status != http.HTTPStatus.NO_CONTENT:
            raise errors.DiscordApiError(resp.data.decode("utf-8"))

    def init_commands(self, app: Flask):
        """Sync the :class:`Command` s configured with this :class:`Discord` instance to the Discord application indicated by the DISCORD_CLIENT_ID key.

        Args
            app: A :class:`Flask` instance configured with a `DISCORD_CLIENT_ID` and `DISCORD_CLIENT_SECRET`
        """
        self.client_id = app.config.setdefault("DISCORD_CLIENT_ID", "")
        self.client_secret = app.config.setdefault("DISCORD_CLIENT_SECRET", "")

        if not (self.client_id and self.client_secret):
            raise ValueError(
                "You must define DISCORD_CLIENT_ID and DISCORD_CLIENT_SECRET configuration values"
            )

        if not self.commands:
            logger.warning(
                "Running init_commands with no commands defined!\n"
                "If you would like discord-interactions-flask to automatically push commands to Discord you will "
                "need to run `init_commands` _after_ defining your commands",
                self,
            )
            return

        self._refresh_token()

        for guild_id, commands in self.commands.items():
            iteractions = self._create_commands(commands.values(), guild_id)

            for command, interaction in zip(commands.values(), iteractions):
                self.runtime_commands[interaction.id] = command



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['discord_interactions_flask']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.1.3,<3.0.0',
 'PyNaCl>=1.5.0,<2.0.0',
 'jsons>=1.6.3,<2.0.0',
 'urllib3>=1.26.10,<2.0.0']

setup_kwargs = {
    'name': 'discord-interactions-flask',
    'version': '0.1.1',
    'description': '',
    'long_description': '# discord-interactions-flask\n\nA [Flask](https://github.com/pallets/flask/) extension to support interacting with [Discord Interactions](https://discord.com/developers/docs/interactions/application-commands).\n\nCheck out the [quickstart](https://docs.davidbuckley.ca/discord-interactions-flask/usage/quickstart.html) or the [examples directory](/examples) for an idea of how to use it.\n\n```python\nimport os\n\nfrom flask import Flask\n\nfrom discord_interactions_flask import Discord\nfrom discord_interactions_flask import helpers\nfrom discord_interactions_flask.interactions ChatInteraction\n\napp = Flask(__name__)\napp.config[\'DISCORD_PUBLIC_KEY\'] = os.environ[\'DISCORD_PUBLIC_KEY\']\napp.config[\'DISCORD_CLIENT_ID\'] = os.environ[\'DISCORD_CLIENT_ID\']\napp.config[\'DISCORD_CLIENT_SECRET\'] = os.environ[\'DISCORD_CLIENT_SECRET\']\n\ndiscord = Discord()\n\n@discord.command("slash-example")\ndef chat_command(interaction: ChatInteraction) -> types.InteractionResponse:\n    return helpers.content_response("Hello, World!")\n\nchat_command.description = "Say hello via a slash command"\n\ndiscord.init_app(app)\n```\n',
    'author': 'David Buckley',
    'author_email': 'david@davidbuckley.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/buckley-w-david/discord-interactions-flask',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from functools import wraps

from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey


def verify_key(
    raw_body: bytes, signature: str, timestamp: str, client_public_key: str
) -> bool:
    message = timestamp.encode() + raw_body
    try:
        vk = VerifyKey(bytes.fromhex(client_public_key))
        vk.verify(message, bytes.fromhex(signature))
        return True
    except BadSignatureError:
        return False


def verify_key_decorator(client_public_key):
    from flask import request

    def _decorator(f):
        @wraps(f)
        def __decorator(*args, **kwargs):
            # Verify request
            signature = request.headers.get("X-Signature-Ed25519")
            timestamp = request.headers.get("X-Signature-Timestamp")
            if (
                signature is None
                or timestamp is None
                or not verify_key(request.data, signature, timestamp, client_public_key)
            ):
                return "Bad request signature", 401

            # Pass through
            return f(*args, **kwargs)

        return __decorator

    return _decorator



"""Catchall for the various datatypes defined in the Discord documentation."""
from typing import Optional, Union, List, Literal, Dict, Any
from dataclasses import dataclass
import enum

from discord_interactions_flask.jsons import BaseModel

Snowflake = str


class InteractionType(enum.IntEnum):
    """Types of interactions supported by Discord."""

    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3
    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    MODAL_SUBMIT = 5


class InteractionCallbackType(enum.IntEnum):
    """Types of responses that can be returned by an interaction webhook."""

    PONG = 1  # 	ACK a Ping
    CHANNEL_MESSAGE_WITH_SOURCE = 4  # 	respond to an interaction with a message
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5  # 	ACK an interaction and edit a response later, the user sees a loading state
    DEFERRED_UPDATE_MESSAGE = 6  # 	for components, ACK an interaction and edit the original message later; the user does not see a loading state
    UPDATE_MESSAGE = (
        7  # 	for components, edit the message the component was attached to
    )
    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = (
        8  # 	respond to an autocomplete interaction with suggested choices
    )
    MODAL = 9  # 	respond to an interaction with a popup modal


class CommandType(enum.IntEnum):
    CHAT = 1
    USER = 2
    MESSAGE = 3


class ApplicationCommandOptionType(enum.IntEnum):
    SUB_COMMAND = 1  #
    SUB_COMMAND_GROUP = 2  #
    STRING = 3  #
    INTEGER = 4  # 	Any integer between -2^53 and 2^53
    BOOLEAN = 5  #
    USER = 6  #
    CHANNEL = 7  # 	Includes all channel types + categories
    ROLE = 8  #
    MENTIONABLE = 9  # 	Includes users and roles
    NUMBER = 10  # 	Any double between -2^53 and 2^53
    ATTACHMENT = 11  # 	attachment object


class ChannelType(enum.IntEnum):
    GUILD_TEXT = 0  # 	a text channel within a server
    DM = 1  # 	a direct message between users
    GUILD_VOICE = 2  # 	a voice channel within a server
    GROUP_DM = 3  # 	a direct message between multiple users
    GUILD_CATEGORY = 4  # 	an organizational category that contains up to 50 channels
    GUILD_NEWS = (
        5  # 	a channel that users can follow and crosspost into their own server
    )
    GUILD_NEWS_THREAD = 10  # 	a temporary sub-channel within a GUILD_NEWS channel
    GUILD_PUBLIC_THREAD = 11  # 	a temporary sub-channel within a GUILD_TEXT channel
    GUILD_PRIVATE_THREAD = 12  # 	a temporary sub-channel within a GUILD_TEXT channel that is only viewable by those invited and those with the MANAGE_THREADS permission
    GUILD_STAGE_VOICE = 13  # 	a voice channel for hosting events with an audience
    GUILD_DIRECTORY = 14  # 	the channel in a hub containing the listed servers
    GUILD_FORUM = 15  # 	(still in development) a channel that can only contain threads


class ComponentType(enum.IntEnum):
    ACTION_ROW = 1  # A container for other components
    BUTTON = 2  # A button object
    SELECT_MENU = 3  # A select menu for picking from choices
    TEXT_INPUT = 4  # A text input object


class ButtonStyle(enum.IntEnum):
    PRIMARY = 1  # 	blurple	custom_id
    SECONDARY = 2  # 	grey	custom_id
    SUCCESS = 3  # 	green	custom_id
    DANGER = 4  # 	red	custom_id
    LINK = 5  # 	grey, navigates to a URL	url


class TextInputStyle(enum.IntEnum):
    SHORT = 1  # A single-line input
    PARAGRAPH = 2  # A multi-line input


class MessageFlags(enum.IntEnum):
    CROSSPOSTED = (
        1 << 0
    )  # this message has been published to subscribed channels (via Channel Following)
    IS_CROSSPOST = (
        1 << 1
    )  # this message originated from a message in another channel (via Channel Following)
    SUPPRESS_EMBEDS = 1 << 2  # do not include any embeds when serializing this message
    SOURCE_MESSAGE_DELETED = (
        1 << 3
    )  # the source message for this crosspost has been deleted (via Channel Following)
    URGENT = 1 << 4  # this message came from the urgent message system
    HAS_THREAD = (
        1 << 5
    )  # this message has an associated thread, with the same id as the message
    EPHEMERAL = (
        1 << 6
    )  # this message is only visible to the user who invoked the Interaction
    LOADING = (
        1 << 7
    )  # this message is an Interaction Response and the bot is "thinking"
    FAILED_TO_MENTION_SOME_ROLES_IN_THREAD = (
        1 << 8
    )  # this message failed to mention some roles and add their members to the thread


class MessageType(enum.IntEnum):
    DEFAULT = 0
    RECIPIENT_ADD = 1
    RECIPIENT_REMOVE = 2
    CALL = 3
    CHANNEL_NAME_CHANGE = 4
    CHANNEL_ICON_CHANGE = 5
    CHANNEL_PINNED_MESSAGE = 6
    USER_JOIN = 7
    GUILD_BOOST = 8
    GUILD_BOOST_TIER_1 = 9
    GUILD_BOOST_TIER_2 = 10
    GUILD_BOOST_TIER_3 = 11
    CHANNEL_FOLLOW_ADD = 12
    GUILD_DISCOVERY_DISQUALIFIED = 14
    GUILD_DISCOVERY_REQUALIFIED = 15
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING = 16
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING = 17
    THREAD_CREATED = 18
    REPLY = 19
    CHAT_INPUT_COMMAND = 20
    THREAD_STARTER_MESSAGE = 21
    GUILD_INVITE_REMINDER = 22
    CONTEXT_MENU_COMMAND = 23
    AUTO_MODERATION_ACTION = 24


# TODO :s/dict/RealType/g


# TODO: This should probably be defined farther down
@dataclass
class Message(BaseModel):
    id: Snowflake  # id of the message
    channel_id: Snowflake  # id of the channel the message was sent in
    author: dict  # user object the author of this message (not guaranteed to be a valid user, see below)
    content: str  # contents of the message
    timestamp: str  # ISO8601 timestamp when this message was sent
    edited_timestamp: Optional[
        str
    ]  # ISO8601 timestamp when this message was edited (or null if never)
    tts: bool  # whether this was a TTS message
    mention_everyone: bool  # whether this message mentions everyone
    mentions: List[
        dict
    ]  # array of user objects   users specifically mentioned in the message
    mention_roles: List[
        str
    ]  # array of role object ids   roles specifically mentioned in this message
    attachments: List[dict]  # array of attachment objects # any attached files
    embeds: List[dict]  # array of embed objects # any embedded content
    pinned: bool  # whether this message is pinned
    type: MessageType  # type of message
    mention_channels: Optional[
        List[dict]
    ] = None  # array of channel mention objects # channels specifically mentioned in this message
    reactions: Optional[
        List[dict]
    ] = None  # array of reaction objects # reactions to the message
    nonce: Optional[Union[int, str]] = None  # used for validating a message was sent
    webhook_id: Optional[
        Snowflake
    ] = None  # if the message is generated by a webhook, this is the webhook's id
    activity: Optional[
        dict
    ] = None  # message activity object # sent with Rich Presence-related chat embeds
    application: Optional[
        dict
    ] = None  # partial application object # sent with Rich Presence-related chat embeds
    application_id: Optional[
        Snowflake
    ] = None  # if the message is an Interaction or application-owned webhook, this is the id of the application
    message_reference: Optional[
        dict
    ] = None  # message reference object # data showing the source of a crosspost, channel follow add, pin, or reply message
    flags: Optional[int] = None  # message flags combined as a bitfield
    referenced_message: Optional[
        "Message"
    ] = None  # message object # the message associated with the message_reference
    interaction: Optional[
        "MessageInteraction"
    ] = None  # message interaction object # sent if the message is a response to an Interaction
    thread: Optional[
        dict
    ] = None  # channel object # the thread that was started from this message, includes thread member object
    components: Optional[
        List["Component"]
    ] = None  # Array of message components # sent if the message contains components like buttons, action rows, or other interactive components
    sticker_items: Optional[
        List[dict]
    ] = None  # array of message sticker item objects # sent if the message contains stickers
    stickers: Optional[
        List[dict]
    ] = None  # array of sticker objects # Deprecated the stickers sent with the message


@dataclass
class Emoji(BaseModel):
    id: Snowflake  # snowflake	emoji id
    name: str  # string (can be null only in reaction emoji objects)	emoji name
    roles: Optional[
        List[Snowflake]
    ]  # array of role object ids	roles allowed to use this emoji
    user: Optional[dict]  # user object	user that created this emoji
    require_colons: Optional[
        bool
    ]  # boolean	whether this emoji must be wrapped in colons
    managed: Optional[bool]  # boolean	whether this emoji is managed
    animated: Optional[bool]  # boolean	whether this emoji is animated
    available: Optional[
        bool
    ]  # boolean	whether this emoji can be used, may be false due to loss of Server Boosts


@dataclass
class ApplicationCommandOptionChoice(BaseModel):
    name: str  # string	1-100 character choice name
    value: Union[
        str, int, float
    ]  # string, integer, or double *	Value for the choice, up to 100 characters if string
    name_localizations: Optional[
        dict
    ] = None  # dictionary with keys in available locales	Localization dictionary for the name field. Values follow the same restrictions as name


@dataclass
class Resolved(BaseModel):
    users: Optional[
        Dict[Snowflake, Any]
    ] = None  # Map of Snowflakes to user objects	the ids and User objects
    members: Optional[
        Dict[Snowflake, Any]
    ] = None  # Map of Snowflakes to partial member objects	the ids and partial Member objects
    roles: Optional[
        Dict[Snowflake, Any]
    ] = None  # Map of Snowflakes to role objects	the ids and Role objects
    channels: Optional[
        Dict[Snowflake, Any]
    ] = None  # Map of Snowflakes to partial channel objects	the ids and partial Channel objects
    messages: Optional[
        Dict[Snowflake, Message]
    ] = None  # Map of Snowflakes to partial messages objects	the ids and partial Message objects
    attachments: Optional[
        Dict[Snowflake, Any]
    ] = None  # Map of Snowflakes to attachment objects	the ids and attachment objects


@dataclass
class ApplicationCommandInteractionDataOption(BaseModel):
    name: str  # string	Name of the parameter
    type: ApplicationCommandOptionType  # integer	Value of application command option type
    value: Optional[
        Union[str, int, float]
    ] = None  # string, integer, or double	Value of the option resulting from user input
    options: Optional[
        List["ApplicationCommandInteractionDataOption"]
    ] = None  # array of application command interaction data option	Present if this option is a group or subcommand
    focused: Optional[
        bool
    ] = None  # boolean	true if this option is the currently focused option for autocomplete


@dataclass
class ApplicationCommandOption(BaseModel):
    type: ApplicationCommandOptionType  # one of application command option type	Type of option
    name: str  # string	1-32 character name
    description: str = ""  # string	1-100 character description
    name_localizations: Optional[
        dict
    ] = None  # dictionary with keys in available locales	Localization dictionary for the name field. Values follow the same restrictions as name
    description_localizations: Optional[
        dict
    ] = None  # dictionary with keys in available locales	Localization dictionary for the description field. Values follow the same restrictions as description
    required: Optional[
        bool
    ] = None  # boolean	If the parameter is required or optional--default false
    choices: Optional[
        List[ApplicationCommandOptionChoice]
    ] = None  # array of application command option choice	Choices for STRING, INTEGER, and NUMBER types for the user to pick from, max 25
    options: Optional[
        List["ApplicationCommandOption"]
    ] = None  # array of application command option	If the option is a subcommand or subcommand group type, these nested options will be the parameters
    channel_types: Optional[
        List[ChannelType]
    ] = None  # array of channel types	If the option is a channel type, the channels shown will be restricted to these types
    min_value: Optional[
        Union[int, float]
    ] = None  # integer for INTEGER options, double for NUMBER options	If the option is an INTEGER or NUMBER type, the minimum value permitted
    max_value: Optional[
        Union[int, float]
    ] = None  # integer for INTEGER options, double for NUMBER options	If the option is an INTEGER or NUMBER type, the maximum value permitted
    autocomplete: Optional[
        bool
    ] = None  # boolean	If autocomplete interactions are enabled for this STRING, INTEGER, or NUMBER type option


@dataclass
class SelectOption(BaseModel):
    label: str  # string	the user-facing name of the option, max 100 characters
    value: str  # string	the dev-defined value of the option, max 100 characters
    description: Optional[
        str
    ] = None  # string	an additional description of the option, max 100 characters
    emoji: Optional[Emoji] = None  # partial emoji object	id, name, and animated
    default: Optional[
        bool
    ] = None  # boolean	will render this option as selected by default


@dataclass
class MessageComponent(BaseModel):
    custom_id: str  # string	the custom_id of the component
    component_type: int  # integer	the type of the component
    values: Optional[
        List[str]
    ]  # array of select option values	values the user selected in a select menu component


@dataclass
class ModalSubmit(BaseModel):
    custom_id: str  # string	the custom_id of the modal
    components: List[
        MessageComponent
    ]  # array of message components	the values submitted by the user


@dataclass
class InteractionData(BaseModel):
    id: Snowflake  # snowflake	the ID of the invoked command
    name: str  # string	the name of the invoked command
    type: CommandType  # integer	the type of the invoked command
    resolved: Optional[
        Resolved
    ] = None  # resolved data	converted users + roles + channels + attachments
    options: Optional[
        List[ApplicationCommandInteractionDataOption]
    ] = None  # array of application command interaction data option	the params + values from the user
    guild_id: Optional[
        Snowflake
    ] = None  # snowflake	the id of the guild the command is registered to
    target_id: Optional[
        Snowflake
    ] = None  # snowflake	id of the user or message targeted by a user or message command


# This structure does not map to PING requests, due to `data`
@dataclass
class Interaction(BaseModel):
    id: Snowflake  # 	snowflake	ID of the interaction
    application_id: Snowflake  # 	snowflake	ID of the application this interaction is for
    type: InteractionType  # 	interaction type	Type of interaction
    token: str  # 	string	Continuation token for responding to the interaction
    version: int  # 	integer	Read-only property, always 1
    data: Union[
        InteractionData, MessageComponent, ModalSubmit
    ]  # 	interaction data	Interaction data payload
    guild_id: Optional[
        Snowflake
    ] = None  # 	snowflake	Guild that the interaction was sent from
    channel_id: Optional[
        Snowflake
    ] = None  # 	snowflake	Channel that the interaction was sent from
    member: Optional[
        dict
    ] = None  # 	guild member object	Guild member data for the invoking user, including permissions
    user: Optional[
        dict
    ] = None  # 	user object	User object for the invoking user, if invoked in a DM
    message: Optional[
        Message
    ] = None  # 	message object	For components, the message they were attached to
    app_permissions: Optional[
        str
    ] = None  # 	string	Bitwise set of permissions the app or bot has within the channel the interaction was sent from
    locale: Optional[str] = None  # 	string	Selected language of the invoking user
    guild_locale: Optional[
        str
    ] = None  # 	Guild's preferred locale, if invoked in a guild


# > This is sent on the message object when the message is a response to an Interaction without an existing message.
# WTF is that word soup
# This is the object that comes in the message.interaction field


@dataclass
class MessageInteraction(BaseModel):
    id: Snowflake  # snowflake	ID of the interaction
    name: str  # string	Name of the application command, including subcommands and subcommand groups
    type: InteractionType  # interaction type	Type of interaction
    user: dict  # user object	User who invoked the interaction
    member: Optional[
        dict
    ]  # partial member object	Member who invoked the interaction in the guild


@dataclass
class Button(BaseModel):
    style: ButtonStyle  # integer	one of button styles
    custom_id: str  # string	a developer-defined identifier for the button, max 100 characters
    label: str  # string	text that appears on the button, max 80 characters
    type: Literal[ComponentType.BUTTON] = ComponentType.BUTTON  # integer	2 for a button
    emoji: Optional[Emoji] = None  # partial emoji	name, id, and animated
    url: Optional[str] = None  # string	a url for link-style buttons
    disabled: bool = False  # boolean	whether the button is disabled (default false)


@dataclass
class SelectMenu(BaseModel):
    custom_id: str  # string	a developer-defined identifier for the select menu, max 100 characters
    options: List[
        SelectOption
    ]  # array of select options	the choices in the select, max 25
    type: Literal[
        ComponentType.SELECT_MENU
    ] = ComponentType.SELECT_MENU  # integer	3 for a select menu
    placeholder: Optional[
        str
    ] = None  # string	custom placeholder text if nothing is selected, max 150 characters
    min_values: Optional[
        int
    ] = None  # integer	the minimum number of items that must be chosen; default 1, min 0, max 25
    max_values: Optional[
        int
    ] = None  # integer	the maximum number of items that can be chosen; default 1, max 25
    disabled: bool = False  # boolean	disable the select, default false


@dataclass
class TextInput(BaseModel):
    custom_id: str  # string	a developer-defined identifier for the input, max 100 characters
    style: TextInputStyle  # integer	the Text Input Style
    label: str  # string	the label for this component, max 45 characters
    type: Literal[
        ComponentType.TEXT_INPUT
    ] = ComponentType.TEXT_INPUT  # integer	4 for a text input
    min_length: Optional[
        int
    ] = None  # integer	the minimum input length for a text input, min 0, max 4000
    max_length: Optional[
        int
    ] = None  # integer	the maximum input length for a text input, min 1, max 4000
    required: bool = (
        True  # boolean	whether this component is required to be filled, default true
    )
    value: Optional[
        str
    ] = None  # string	a pre-filled value for this component, max 4000 characters
    placeholder: Optional[
        str
    ] = None  # string	custom placeholder text if the input is empty, max 100 characters


InteractiveComponent = Union[Button, SelectMenu, TextInput]


@dataclass
class ActionRow(BaseModel):
    components: List[InteractiveComponent]
    type: Literal[
        ComponentType.ACTION_ROW
    ] = ComponentType.ACTION_ROW  # integer	1 for a action row


Component = Union[ActionRow, InteractiveComponent]


@dataclass
class Modal(BaseModel):
    custom_id: str  # a developer-defined identifier for the component, max 100 characters
    title: str  # the title of the popup modal, max 45 characters
    components = List[
        TextInput
    ]  # between 1 and 5 (inclusive) components that make up the modal


@dataclass
class InteractionCallbackDataMessage(BaseModel):
    tts: Optional[bool] = None  # boolean	is the response TTS
    content: Optional[str] = None  # string	message content
    embeds: Optional[List[dict]] = None  # array of embeds	supports up to 10 embeds
    allowed_mentions: Optional[dict] = None  # allowed mentions	allowed mentions object
    flags: Optional[
        int
    ] = None  # integer	message flags combined as a bitfield (only SUPPRESS_EMBEDS and EPHEMERAL can be set)
    components: Optional[
        List[ActionRow]
    ] = None  # array of components	message components
    attachments: Optional[
        List[dict]
    ] = None  # array of partial attachment objects	attachment objects with filename and description


@dataclass
class InteractionCallbackDataAutocomplete(BaseModel):
    choices: List[ApplicationCommandOptionChoice]


@dataclass
class InteractionCallbackDataModal(BaseModel):
    custom_id: str  # string	a developer-defined identifier for the component, max 100 characters
    title: str  # string	the title of the popup modal, max 45 characters
    components: List[
        Component
    ]  # array of components	between 1 and 5 (inclusive) components that make up the modal


InteractionCallbackData = Union[
    InteractionCallbackDataMessage,
    InteractionCallbackDataAutocomplete,
    InteractionCallbackDataModal,
]


@dataclass
class InteractionResponse(BaseModel):
    type: InteractionCallbackType
    data: Optional[InteractionCallbackData] = None


# This is what is sent back to us when we create a command
@dataclass
class ApplicationCommand(BaseModel):
    id: Snowflake  # snowflake	Unique ID of command	all
    application_id: Snowflake  # snowflake	ID of the parent application	all
    name: str  # string	Name of command, 1-32 characters	all
    description: str  # string	Description for CHAT_INPUT commands, 1-100 characters. Empty string for USER and MESSAGE commands	all
    version: Snowflake  # snowflake	Autoincrementing version identifier updated during substantial record changes	all
    type: CommandType = (
        CommandType.CHAT
    )  # one of application command type	Type of command, defaults to 1	all
    guild_id: Optional[
        Snowflake
    ] = None  # snowflake	guild id of the command, if not global	all
    name_localizations: Optional[
        dict
    ] = None  # dictionary with keys in available locales	Localization dictionary for name field. Values follow the same restrictions as name	all
    description_localizations: Optional[
        dict
    ] = None  # dictionary with keys in available locales	Localization dictionary for description field. Values follow the same restrictions as description	all
    options: Optional[
        List[ApplicationCommandOption]
    ] = None  # array of application command option	Parameters for the command, max of 25	CHAT_INPUT
    default_member_permissions: Optional[
        str
    ] = None  # string	Set of permissions represented as a bit set	all
    dm_permission: Optional[
        bool
    ] = None  # boolean	Indicates whether the command is available in DMs with the app, only for globally-scoped commands. By default, commands are visible.	all
    default_permission: Optional[
        bool
    ] = None  # boolean	Not recommended for use as field will soon be deprecated. Indicates whether the command is enabled by default when the app is added to a guild, defaults to true	all
